#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import decimal

import os_resource_classes as orc
import testtools

from placement.objects import usage


class TestUsageNoDB(testtools.TestCase):
    def test_decimal_to_int(self):
        dmal = decimal.Decimal('10')
        usage_obj = usage.Usage(resource_class=orc.VCPU, usage=dmal)
        # Type must come second in assertIsInstance.
        self.assertIsInstance(usage_obj.usage, int)
